/*
  cpumunch - start a bunch of threads that perform useless calculations

  Copyright (C) 2014 Katie Stafford (katie@ktpanda.org)

  Permission is hereby granted, free of charge, to any person obtaining
  a copy of this software and associated documentation files (the
  "Software"), to deal in the Software without restriction, including
  without limitation the rights to use, copy, modify, merge, publish,
  distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to
  the following conditions:

  The above copyright notice and this permission notice shall be
  included in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/


#ifdef WIN32
#include <windows.h>
#include <process.h>
#else
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/signal.h>
#endif

#include <stdio.h>
#include <mmintrin.h> /* MMX instrinsics  __m64 integer type  */
#include <xmmintrin.h> /* SSE  __m128  float */

#ifdef WIN32
/* On Windows x86, new threads are not required to have a stack
 * aligned to 16 bytes, which is required for SSE. This attribute
 * forces GCC to emit code to align the stack. */
void __attribute((force_align_arg_pointer)) munch_sse(void* data)
#else
void* munch_sse(void* data)
#endif
{
    /* Perform a bunch of useless SSE2 operations */
    __m128 m1 = {1.0, 2.0, 3.0, 4.0};
    __m128 m2 = {-1.0, -2.0, -3.0, -4.0};
    __m128 m3 = {2.0, 256.0, 65536.0, 4294967296.0};
    __m128 m4 = {1.0/2.0, 1.0/256.0, 1.0/65536.0, 1.0/4294967296.0};
    __m128 dummy;

    while (1) {
        dummy = _mm_add_ps(dummy, m1);
        dummy = _mm_add_ps(dummy, m2);
        dummy = _mm_mul_ps(dummy, m3);
        dummy = _mm_mul_ps(dummy, m4);
    }
}

void* munch_null(void* data) {
    while (1) { }
}

#ifdef WIN32

void start_thread(void) {
    _beginthread(munch_sse, 0, NULL);
}

#else
void start_thread(void) {
    pthread_t thr;
    pthread_attr_t thrattr;

    pthread_attr_init(&thrattr);

    pthread_attr_setdetachstate(&thrattr, PTHREAD_CREATE_DETACHED);
    if (pthread_create(&thr, &thrattr, munch_sse, NULL)) {
        perror("pthread_create");
    }
}
#endif

int main(int argc, char** argv) {
    int i, numthreads, numcores;

#ifdef WIN32
    SYSTEM_INFO sysinfo;
    GetSystemInfo(&sysinfo);
    numcores = sysinfo.dwNumberOfProcessors;
#else
    numcores = sysconf(_SC_NPROCESSORS_ONLN);
#endif
    printf("%d cores detected\n", numcores);

    if (argc > 1) {
        numthreads = atoi(argv[1]);
    } else {
        numthreads = numcores;
    }

    if (numthreads < 1)
        numthreads = 1;

    if (numthreads > numcores * 16)
        numthreads = numcores * 16;

    printf("Munching with %d threads, press Ctrl-C to stop\n", numthreads);

    for (i = 0; i < numthreads - 1; i++)
      start_thread();

    munch_sse(NULL);

    return 0;
}
